@@ -30,6 +30,14 @@
 #include "luks.h"
 #include "config.h"
 
+/* Enable autodetection if possible */
+#ifdef HAVE_BLKID
+#include <blkid/blkid.h>
+#endif
+
+/* Using our private realpath function */
+#include "realpath.h"
+
 /* error codes */
 const int E_ARGS = 1;
 const int E_DEVICE = 2;
@@ -41,6 +49,9 @@ const int E_PID = 7;
 const int E_LOCKED = 8;
 const int E_INTERNAL = 100;
 
+#define OPT_FMASK 128
+#define OPT_DMASK 129
+
 /**
  * Print some help.
  * @param exename Name of the executable (argv[0]).
@@ -51,7 +62,7 @@ usage( const char* exename )
     printf( _("Usage:\n\n%s [options] <device> [<label>]\n\n"
     "  Mount <device> to a directory below %s if policy requirements\n"
     "  are met (see pmount(1) for details). If <label> is given, the mount point\n"
-    "  will be %s/<label>, otherwise it will be %s<device>.\n"
+    "  will be %s<label>, otherwise it will be %s<device>.\n"
     "  If the mount point does not exist, it will be created.\n\n"),
         exename, MEDIADIR, MEDIADIR, MEDIADIR );
 
@@ -69,19 +80,25 @@ usage( const char* exename )
     "  -r          : force <device> to be mounted read-only\n"
     "  -w          : force <device> to be mounted read-write\n"
     "  -s, --sync  : mount <device> with the 'sync' option (default: 'async')\n"
-    "  --noatime   : mount <device> with the 'noatime' option (default: 'atime')\n"
+    "  -A, --noatime\n"
+    "                mount <device> with the 'noatime' option (default: 'atime')\n"
     "  -e, --exec  : mount <device> with the 'exec' option (default: 'noexec')\n"
     "  -t <fs>     : mount as file system type <fs> (default: autodetected)\n"
     "  -c <charset>: use given I/O character set (default: 'utf8' if called\n"
     "                in an UTF-8 locale, otherwise mount default)\n"
     "  -u <umask>  : use specified umask instead of the default (only for\n"
     "                file sytems which actually support umask setting)\n"
-    " --passphrase <file>\n"
+    "  --fmask <fmask>\n"
+    "                use specified fmask\n"
+    "  --dmask <dmask>\n"
+    "                use specified dmask\n"
+    "  -p <file>, --passphrase <file>\n"
     "                read passphrase from file instead of the terminal\n"
     "                (only for LUKS encrypted devices)\n"
     "  -d, --debug : enable debug output (very verbose)\n"
-    "  -h, --help  : print help message and exit successfuly\n"
-    "  --version   : print version number and exit successfully") );
+    "  -h, --help  : print this help message and exit successfuly\n"
+    "  -V, --version\n"
+    "                print version number and exit successfully") );
 }
 
 /**
@@ -201,24 +218,32 @@ do_mount_fstab( const char* device )
  * @param force_write 1 for forced r/w, 0 for forced r/o, -1 for kernel default
  * @param iocharset charset to use for file name conversion; NULL for mount
  *        default
+ * @param utf8 is true if the option utf8 should be used for VFAT
  * @param umask User specified umask (NULL for default)
+ * @param fmask User specified fmask (NULL for umask)
+ * @param dmask User specified dmask (NULL for umask)
  * @param suppress_errors: if true, stderr is redirected to /dev/null
  * @return exit status of mount, or -1 on failure.
  */
 int
 do_mount( const char* device, const char* mntpt, const char* fsname, int async,
-        int noatime, int exec, int force_write, const char* iocharset, const
-        char* umask, int suppress_errors )
+	  int noatime, int exec, int force_write, const char* iocharset, 
+	  int utf8, 
+	  const char* umask, const char *fmask, const char *dmask, 
+	  int suppress_errors )
 {
     const struct FS* fs;
     char ugid_opt[100];
     char umask_opt[100];
+    char fdmask_opt[100];
     char iocharset_opt[100];
     const char* sync_opt = ",sync";
     const char* atime_opt = ",atime";
     const char* exec_opt = ",noexec";
     const char* access_opt = NULL;
     char options[1000];
+    /* We deal with masks in another way now: */
+    unsigned i_umask, i_dmask, i_fmask;
 
     /* check and retrieve option information for requested file system */
     if( !fsname) {
@@ -232,14 +257,24 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
         return -1;
     }
 
-    /* validate user specified umask */
+    /* validate user specified masks */
     if( umask && parse_unsigned( umask, E_ARGS ) > 0777 ) {
         fprintf( stderr, _("Error: invalid umask %s\n"), umask );
         return -1;
     }
 
+    if( fmask && parse_unsigned( fmask, E_ARGS ) > 0777 ) {
+        fprintf( stderr, _("Error: invalid fmask %s\n"), fmask );
+        return -1;
+    }
+
+    if( dmask && parse_unsigned( dmask, E_ARGS ) > 0777 ) {
+        fprintf( stderr, _("Error: invalid dmask %s\n"), dmask );
+        return -1;
+    }
+
     /* assemble option string */
-    *ugid_opt = *umask_opt = *iocharset_opt = 0;
+    *ugid_opt = *umask_opt = *fdmask_opt = *iocharset_opt = 0;
     if( fs->support_ugid ) {
 	struct stat statbuf;
 	int gid = getgid();
@@ -261,8 +296,31 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
     }
 
     if( fs->umask )
-        snprintf( umask_opt, sizeof( umask_opt ), ",umask=%s", 
-                umask ? umask : fs->umask );
+      snprintf( umask_opt, sizeof( umask_opt ), ",umask=%s", 
+		umask ? umask : fs->umask );
+    /* If the fs supports fdmasks, we try to make some values
+       up.
+    */
+    if( fs->umask && fs->fdmask ) {
+      /* We first get the umask value */
+      if(umask)
+	i_umask = parse_unsigned( umask, E_ARGS ); /* shouldn't fail */
+      else
+	i_umask = parse_unsigned( fs->umask, E_ARGS ); /* shouldn't fail */
+
+      /* Now the fmask */
+      if(fmask)
+	i_fmask = parse_unsigned( fmask, E_ARGS ); /* shouldn't fail */
+      else			/* make up from the umask parameter */
+	i_fmask = i_umask | 0111; /* remove exec permissions */
+      /* And the dmask */
+      if(dmask)
+	i_dmask = parse_unsigned( dmask, E_ARGS ); /* shouldn't fail */
+      else			/* make up from the umask parameter */
+	i_dmask = i_umask;	/* same as umask */
+      snprintf( fdmask_opt, sizeof( fdmask_opt ), fs->fdmask, 
+		i_fmask, i_dmask );
+    }
 
     if( async )
         sync_opt = ",async";
@@ -280,17 +338,42 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
     else
         access_opt = "";
 
-    if( iocharset && fs->support_iocharset ) {
+    if( iocharset && fs->iocharset_format ) {
         if( !is_word_str( iocharset ) ) {
             fprintf( stderr, _("Error: invalid charset name '%s'\n"), iocharset );
             return -1;
         }
-        snprintf( iocharset_opt, sizeof( iocharset_opt ), ",iocharset=%s", iocharset );
+	/* VFAT and UTF-8 need special care, see bug #443514 and mount(1) */
+	if(! strcmp(fsname, "vfat") && utf8 ) {
+	  debug("VFAT in a UTF-8 locale: using option utf8\n");
+	  if(! strcmp(iocharset, "utf8")) {
+	    debug("filesystem is vfat and charset is utf-8: using iso8859-1\n"
+		  "You can change with the -c option");
+	    snprintf( iocharset_opt, sizeof( iocharset_opt ), 
+		      ",utf8,iocharset=iso8859-1" );
+	  }
+	  else {
+	    snprintf( iocharset_opt, sizeof( iocharset_opt ), 
+		      ",utf8,iocharset=%s", iocharset );
+
+	  }
+	} else {
+	  snprintf( iocharset_opt, sizeof( iocharset_opt ), 
+		  fs->iocharset_format, iocharset );
+	}
+    }
+    else if(! strcmp(fsname, "vfat") && fs->iocharset_format) {
+      /* We still make a special case for vfat, as in certain cases,
+	 mount will mount it with iocharset=utf8, some times without
+	 warning. So, in the absence of a specified charset, we
+	 force iocharset=iso8859-1*/
+      snprintf( iocharset_opt, sizeof( iocharset_opt ), 
+		fs->iocharset_format, "iso8859-1");
     }
 
-    snprintf( options, sizeof( options ), "%s%s%s%s%s%s%s%s", 
+    snprintf( options, sizeof( options ), "%s%s%s%s%s%s%s%s%s", 
             fs->options, sync_opt, atime_opt, exec_opt, access_opt, ugid_opt,
-            umask_opt, iocharset_opt );
+            umask_opt, fdmask_opt, iocharset_opt );
 
     /* go for it */
     return spawnl( SPAWN_EROOT | SPAWN_RROOT | (suppress_errors ? SPAWN_NO_STDERR : 0 ),
@@ -310,33 +393,68 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
  * @param force_write 1 for forced r/w, 0 for forced r/o, -1 for kernel default
  * @param iocharset charset to use for file name conversion; NULL for mount
  *        default
+ * @param utf8 is true if the option utf8 should be used for VFAT
  * @param umask User specified umask (NULL for default)
+ * @param fmask User specified fmask (NULL for umask)
+ * @param dmask User specified dmask (NULL for umask)
  * @return last return value of do_mount (i. e. 0 on success, != 0 on error)
  */
 int
 do_mount_auto( const char* device, const char* mntpt, int async, 
-        int noatime, int exec, int force_write, const char* iocharset, 
-        const char* umask )
+	       int noatime, int exec, int force_write, const char* iocharset, 
+	       int utf8, 
+	       const char* umask, const char *fmask, const char *dmask )
 {
     const struct FS* fs;
     int nostderr = 1;
     int result = -1;
+    struct stat buf;		/* Not used */
+
+    /* First, if that is supported, we try with blkid */
+#ifdef HAVE_BLKID
+    const char * tp;
+    blkid_cache c;
+    blkid_get_cache(&c, "/dev/null");
+    tp = blkid_get_tag_value(c, "TYPE", device);
+    if(tp) {
+      debug("blkid gave FS %s for '%s'\n", tp, device);
+      if(! strcmp(tp, "ntfs") && !stat(MOUNT_NTFS_3G, &buf)) {
+	debug("blkdid detected ntfs and ntfs-3g was found. Using ntfs-3g\n");
+	tp = "ntfs-3g";
+      }
+      result = do_mount( device, mntpt, tp, async, noatime, exec, 
+			 force_write, iocharset, utf8, umask, fmask, 
+			 dmask, nostderr );
+      if(result == 0)
+	return result;
+      debug("blkid-detected FS failed, trying manually \n");
+    }
+    blkid_put_cache(c);
+#endif /* HAVE_BLKID */
+
+    result = -1;
 
     for( fs = get_supported_fs(); fs->fsname; ++fs ) {
-        /* don't suppress stderr if we try the last possible fs */
-        if( (fs+1)->fsname == NULL )
-            nostderr = 0;
-        result = do_mount( device, mntpt, fs->fsname, async, noatime, exec,
-                force_write, iocharset, umask, nostderr );
-        if( result == 0 )
-            break;
+      /* Skip fs marked as such unless it is ntfs-3g and
+	 we can stat MOUNT_NTFS_G3 
+      */
+      if(fs->skip_autodetect && 
+	 !(! strcmp(fs->fsname, "ntfs-3g") && !stat(MOUNT_NTFS_3G, &buf)))
+	continue;		/* skip fs that are marked as such */
+      /* don't suppress stderr if we try the last possible fs */
+      if( (fs+1)->fsname == NULL )
+	nostderr = 0;
+      result = do_mount( device, mntpt, fs->fsname, async, noatime, exec,
+			 force_write, iocharset, utf8, umask, fmask, dmask, nostderr );
+      if( result == 0 )
+	break;
 
-	/* sometimes VFAT fails when using iocharset; try again without */
-	if( iocharset )
-	    result = do_mount( device, mntpt, fs->fsname, async, noatime, exec,
-                    force_write, NULL, umask, nostderr );
-        if( result <= 0 )
-            break;
+      /* sometimes VFAT fails when using iocharset; try again without */
+      if( iocharset )
+	result = do_mount( device, mntpt, fs->fsname, async, noatime, exec,
+			   force_write, NULL, utf8, umask, fmask, dmask, nostderr );
+      if( result <= 0 )
+	break;
     }
     return result;
 }
@@ -503,7 +621,10 @@ main( int argc, char** argv )
     const char* use_fstype = NULL;
     const char* iocharset = NULL;
     const char* umask = NULL;
+    const char* fmask = NULL;
+    const char* dmask = NULL;
     const char* passphrase = NULL;
+    int utf8 = -1; 		/* Whether we live in a UTF-8 world or not */
     int result;
 
     enum { MOUNT, LOCK, UNLOCK } mode = MOUNT;
@@ -520,6 +641,8 @@ main( int argc, char** argv )
         { "type", 1, NULL, 't' },
         { "charset", 1, NULL, 'c' },
         { "umask", 1, NULL, 'u' },
+        { "fmask", 1, NULL, OPT_FMASK },
+        { "dmask", 1, NULL, OPT_DMASK },
         { "passphrase", 1, NULL, 'p' },
         { "read-only", 0, NULL, 'r' },
         { "read-write", 0, NULL, 'w' },
@@ -532,6 +655,17 @@ main( int argc, char** argv )
     bindtextdomain( "pmount", NULL );
     textdomain( "pmount" );
 
+    /* If pmount is run without a single argument, print out the list
+       of removable devices. Does not require root privileges, just read access
+       to the /proc/mounts file.
+    */
+    if( argc == 1 ) {
+      printf(_("Printing mounted removable devices:\n"));
+      print_mounted_removable_devices();
+      printf(_("To get a short help, run %s -h\n"), argv[0]);
+      return 0;
+    }
+
     /* are we root? */
     if( geteuid() ) {
         fputs( _("Error: this program needs to be installed suid root\n"), stderr );
@@ -541,9 +675,10 @@ main( int argc, char** argv )
     /* drop root privileges until we really need them (still available as saved uid) */
     seteuid( getuid() );
 
+
     /* parse command line options */
     do {
-        switch( option = getopt_long( argc, argv, "+hdelLsArwt:c:u:V", long_opts, NULL ) ) {
+        switch( option = getopt_long( argc, argv, "+hdelLsArwp:t:c:u:V", long_opts, NULL ) ) {
             case -1:  break;          /* end of arguments */
             case ':':
             case '?': return E_ARGS;  /* unknown argument */
@@ -567,6 +702,10 @@ main( int argc, char** argv )
             case 'c': iocharset = optarg; break;
 
             case 'u': umask = optarg; break;
+	    
+	    case OPT_FMASK: fmask = optarg; break;
+	    
+	    case OPT_DMASK: dmask = optarg; break;
 
             case 'p': passphrase = optarg; break;
 
@@ -595,8 +734,11 @@ main( int argc, char** argv )
         return E_ARGS;
     }
 
-    /* if we got a mount point, convert it to a device */
-    if( fstab_has_mntpt( "/etc/fstab", devarg, mntptdev, sizeof(mntptdev) ) ) {
+    /* Lookup in /etc/fstab if devarg is a mount point, unless we already
+       have a block device -- this way, pmount shouldn't choke on stale
+       network mounts. */
+    if( (! is_block(devarg)) && 
+	fstab_has_mntpt( "/etc/fstab", devarg, mntptdev, sizeof(mntptdev) ) ) {
         debug( "resolved mount point %s to device %s\n", devarg, mntptdev );
         devarg = mntptdev;
     }
@@ -611,7 +753,7 @@ main( int argc, char** argv )
     }
 
     /* is the device already handled by fstab? We allow is_real_path == 0 here
-     * to transparently mount thinks like NFS and SMB drives */
+     * to transparently mount things like NFS and SMB drives */
     fstab_device = fstab_has_device( "/etc/fstab", device, NULL, NULL );
     if( mode == MOUNT && fstab_device ) {
         if( arg2 )
@@ -633,6 +775,16 @@ main( int argc, char** argv )
                 return E_DEVICE;
             }
             debug( "trying to prepend '" DEVDIR "' to device argument, now %s\n", device );
+	    /* We need to lookup again in fstab: */
+	    fstab_device = fstab_has_device( "/etc/fstab", device, NULL, NULL );
+	    if( mode == MOUNT && fstab_device ) {
+	      if( arg2 )
+		fprintf( stderr, _("Warning: device %s is already handled by /etc/fstab,"
+				   " supplied label is ignored\n"), fstab_device );
+	      
+	      do_mount_fstab( fstab_device );
+	      return E_EXECMOUNT;
+	    }
         }
     }
 
@@ -662,6 +814,17 @@ main( int argc, char** argv )
                     iocharset = "utf8";
                 }
             }
+	    /* If user did not choose explicitly for or against utf8 */
+	    if( utf8 == -1 ) {
+	      const char* codeset;
+	      codeset = nl_langinfo( CODESET );
+	      if( codeset && !strcmp( codeset, "UTF-8" ) ) {
+		debug( "locale encoding uses UTF-8: will mount FAT with utf8 option" );
+		utf8 = 1;
+	      } else {
+		utf8 = 0;
+	      }
+	    }
 
             /* clean stale locks */
             clean_lock_dir( device );
@@ -682,6 +845,9 @@ main( int argc, char** argv )
                     fprintf( stderr, _("Error: mapped device already exists\n") );
                     exit( E_POLICY );
                 case DECRYPT_OK:
+		  /* We create a luks lockfile _on the decrypted device !_*/
+		  if(! luks_create_lockfile(decrypted_device))
+		    fprintf(stderr, _("Warning: could not create luks lockfile\n"));
                 case DECRYPT_NOTENCRYPTED:
                     break;
                 default:
@@ -701,10 +867,10 @@ main( int argc, char** argv )
             /* off we go */
             if( use_fstype )
                 result = do_mount( decrypted_device, mntpt, use_fstype, async, noatime,
-                        exec, force_write, iocharset, umask, 0 );
+				   exec, force_write, iocharset, utf8, umask, fmask, dmask, 0 );
             else
                 result = do_mount_auto( decrypted_device, mntpt, async, noatime, exec,
-                        force_write, iocharset, umask ); 
+                        force_write, iocharset, utf8, umask, fmask, dmask ); 
 
             /* unlock the mount point again */
             debug( "unlocking mount point directory\n" );
@@ -713,7 +879,7 @@ main( int argc, char** argv )
 
             if( result ) {
                 if( decrypt == DECRYPT_OK )
-                    luks_release( decrypted_device );
+                    luks_release( decrypted_device, 0 ); 
 
                 /* mount failed, delete the mount point again */
                 if( remove_pmount_mntpt( mntpt ) ) {
